/*
 * Description: Trace the open syscall and generates a list of files created
 * by a program, using LD_PRELOAD mechanism.
 * Author: Catalin(ux) M. BOIE
 * E-mail: catab at embedromix dot ro
 * Web: http://kernel.embedromix.ro/us/
 */

#define __USE_GNU
#define	_GNU_SOURCE
#define __USE_XOPEN2K
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <regex.h>

#include "logmagic_config.h"


static int			(*old_open)(const char *path, int flags, ...);
static FILE			*(*old_fopen)(const char *path, const char *mode);
static int			(*old_creat)(const char *path, mode_t mode);

static char			*conf_file = "/etc/logmagic.conf";
static char			trace_file[4096];
static int			Log_fd;
static FILE			*Log;

static unsigned char		inited, disabled, test_match;

static char			regex[1024];
static regex_t			reg;


static void xlog(const char *format, ...)
{
	va_list ap;

	if (Log == NULL)
		return;

	va_start(ap, format);
	vfprintf(Log, format, ap);
	va_end(ap);
}


/* Functions */

static int match(const char *path)
{
	int r;

	/* Match if no re specified */
	if (!test_match)
		return 1;

	r = regexec(&reg, path, 0, NULL, 0 /*flags*/);
	if (r == 0)
		return 1;

	return 0;
}

static void load_conf(void)
{
	char line[4096], path[4096];
	char *s, *p;
	int r;
	unsigned int len;
	FILE *f;

	if (disabled)
		return;

	s = getenv("LOGMAGIC_CONF");
	if (s)
		conf_file = s;

	f = old_fopen(conf_file, "r");
	if (!f)
		f = old_fopen("logmagic.conf", "r");
	if (f) {
		while (!feof(f)) {
			p = fgets(line, sizeof(line), f);
			if (!p)
				break;

			if ((line[0] == '#') || (line[0] == '\n'))
				continue;

			len = strlen(line);
			while ((len > 0)
				&& ((line[len - 1] == '\n')
				|| (line[len - 1] == ' ')
				|| (line[len - 1] == '\t')
				|| (line[len - 1] == '\r')))
				line[--len] = '\0';

			s = strchr(line, '#');
			if (s)
				*s = '\0';

			if (strncmp(line, "trace_file ", 11) == 0) {
				s = line + 11;
				while ((*s == ' ') || (*s == '\t'))
					s++;
				snprintf(trace_file, sizeof(trace_file), "%s", s);
			}

			if (strncmp(line, "regex ", 6) == 0) {
				s = line + 6;
				while ((*s == ' ') || (*s == '\t'))
					s++;
				snprintf(regex, sizeof(regex), "%s", s);
			}
		}

		fclose(f);
	}

	s = getenv("LOGMAGIC_TRACE_FILE");
	if (s)
		snprintf(trace_file, sizeof(trace_file), "logmagic.log");

	snprintf(path, sizeof(path), "%s-%u", trace_file, getuid());
	Log = old_fopen(path, "w+");
	if (!Log) {
		fprintf(stderr, "Cannot open trace file [%s] (%s)!\n",
			path, strerror(errno));
		disabled = 1;
		return;
	}
	setlinebuf(Log);
	Log_fd = fileno(Log);

	s = getenv("LOGMAGIC_REGEX");
	if (s)
		snprintf(regex, sizeof(regex), "%s", s);

	r = regcomp(&reg, regex, REG_EXTENDED | REG_NOSUB);
	if (r != 0) {
		fprintf(stderr, "Invalid regular expression (%d)!\n", r);
		disabled = 1;
		return;
	}
	test_match = 1;
}

static void init(void)
{
	if (inited == 1)
		return;

	/******** hijack system calls ********/

	old_open = dlsym(RTLD_NEXT, "open");
	if (old_open == NULL) {
		fprintf(stderr, "Cannot resolve 'open' function!\n");
		disabled = 1;
		return;
	}

	old_fopen = dlsym(RTLD_NEXT, "fopen");
	if (old_fopen == NULL) {
		fprintf(stderr, "Cannot resolve 'fopen' function!\n");
		disabled = 1;
		return;
	}

	old_creat = dlsym(RTLD_NEXT, "creat");
	if (old_creat == NULL) {
		fprintf(stderr, "Cannot resolve 'creat' function!\n");
		disabled = 1;
		return;
	}

	/* Defaults */
	strcpy(trace_file, "logmagic.log");

	load_conf();

	inited = 1;
}

/*
 * Stores a path if all conditions are met
 */
static void store(const char *path)
{
	struct timeval now;
	char *d;

	if (!path)
		return;

	if (disabled)
		return;

	if (!match(path))
		return;

	if (path[0] == '/') {
		d = NULL;
	} else {
		d = get_current_dir_name();
	}

	gettimeofday(&now, NULL);

	if (flock(Log_fd, LOCK_EX) == -1)
		return;

	xlog("ts=%ld path=%s%s%s\n", now.tv_sec, d ? d : "", d ? "/" : "", path);

	flock(Log_fd, LOCK_UN);

	free(d);
}

int open(const char *path, int flags, ...)
{
	int fd;
	va_list ap;
	mode_t mode;

	init();

	/* We need only files created! */
	if ((flags & O_CREAT) == 0)
		return old_open(path, flags);

	va_start(ap, flags);
	mode = va_arg(ap, mode_t);
	va_end(ap);

	fd = old_open(path, flags, mode);
	if (fd == -1)
		return -1;

	store(path);
	return fd;
}

int creat(const char *path, mode_t mode)
{
	int fd;

	init();

	fd = old_creat(path, mode);
	if (fd == -1)
		return -1;

	store(path);
	return fd;
}

FILE *fopen(const char *path, const char *mode)
{
	FILE *f;

	init();

	/* We need only files created! */
	if (!strchr(mode, 'w') && !strchr(mode, 'a'))
		return old_fopen(path, mode);

	f = old_fopen(path, mode);
	if (!f)
		return NULL;

	store(path);
	return f;
}
